import os
from ufl import (
    TestFunction,
    TrialFunction,
    dx,
    inner,
    inner,
)
from dolfinx.fem import Function
from dolfinx.fem.petsc import LinearProblem
from dolfinx.mesh import locate_entities_boundary, meshtags
import numpy as np



def make_unique_dir(base_path):
    """
    Create a unique directory. If `base_path` exists, appends _1, _2, ... until an unused path is found.

    Returns the final created path.
    """
    dir_path = base_path
    counter = 1

    while os.path.exists(dir_path):
        dir_path = f"{base_path}_{counter}"
        counter += 1

    os.makedirs(dir_path)
    return dir_path

def project(scalarfield, space_CG):
    field = scalarfield.copy()
    w = TrialFunction(space_CG)
    phi = TestFunction(space_CG)
    
    a = inner(w, phi) * dx 
    L = inner(field, phi) * dx 
    
    problem = LinearProblem(a, L, petsc_options={"ksp_type": "preonly", "pc_type": "lu"})
    field_proj = problem.solve()
    
    return field_proj

def extract_values(q, *args):
    Q = q.function_space
    if len(args) == 0:
        dof_map = Q.dofmap.list.flatten()
        return q.x.array[dof_map]
    elif len(args) == 1:
        i = args[0]
        dof_map = Q.sub(i).dofmap.list.flatten()
        return q.sub(i).x.array[dof_map]
    elif len(args) == 2:
        i = args[0]
        j = args[1]
        dof_map = Q.sub(i).sub(j).dofmap.list.flatten()
        return q.sub(i).x.array[dof_map]
    else:
        assert False
        

def generate_facets_tags(domain, P0, P1):
    """
    Generate facet tags for the boundary of the mesh.
    """
    # Get the mesh topology dimension and facet geometrical dimension
    mesh_topology_dim = domain.topology.dim
    facet_geometrical_dimension = mesh_topology_dim - 1
    fdim = facet_geometrical_dimension
    left = locate_entities_boundary(domain, fdim, lambda x: np.isclose(x[0], P0[0]))
    right = locate_entities_boundary(domain, fdim, lambda x: np.isclose(x[0], P1[0]))
    bottom = locate_entities_boundary(domain, fdim, lambda x: np.isclose(x[1], P0[1]))
    top = locate_entities_boundary(domain, fdim, lambda x: np.isclose(x[1], P1[1]))

    facets = np.concatenate([left, right, bottom, top])
    values = np.concatenate([
        np.full_like(left, 1),
        np.full_like(right, 2),
        np.full_like(bottom, 3),
        np.full_like(top, 4),
    ])

    sorted_facets = np.argsort(facets)
    facet_tags = meshtags(domain, fdim, facets[sorted_facets], values[sorted_facets])
    return facet_tags
        
def setup_notebook_paths():
    try:
        CMM_DIR = os.getcwd()
        os.environ["PYTHONPATH"] = f"{CMM_DIR}:{os.environ.get('PYTHONPATH', '')}"
        username = os.getenv("USER") or os.geteuid()

        # Set XDG_RUNTIME_DIR to a valid directory
        os.environ["XDG_RUNTIME_DIR"] = f"/tmp/runtime-{username}"
        os.makedirs(os.environ["XDG_RUNTIME_DIR"], exist_ok=True)
        os.chmod(os.environ["XDG_RUNTIME_DIR"], 0o700)
    except Exception as e:
        print(f"Failed to set XDG_RUNTIME_DIR: {e}")
        print("You might get some warnings, but it should still work.")
        
        