import dolfinx
import ufl
import numpy as np
from dolfinx import fem, mesh as dmesh
from mpi4py import MPI
import basix
from dolfinx.fem.petsc import LinearProblem
import os


CMM_DIR = os.getcwd()
os.environ["PYTHONPATH"] = f"{CMM_DIR}:{os.environ.get('PYTHONPATH', '')}"

try:
    from library.helpers import project
except ImportError:
    from helpers import project


class VertexBasedLimiter:
    """
    Kuzmin-style vertex-based slope limiter for 2D 'DG(1)' fields
    on either triangular or quadrilateral meshes in dolfinx (>= 0.9).
    """

    def __init__(self, space_dg, alpha=1.0):
        """
        Initialize the limiter for a given DG(1) space on a 2D mesh.

        Args:
            space_dg: A dolfinx.fem.FunctionSpace with element
                      ("DG", 1) or ("DQ", 1) on a 2D triangle or quad mesh.
        """
        V = space_dg
        elem = V.ufl_element()

        cellname = elem.cell.cellname()  # e.g. "triangle" or "quadrilateral"
        degree = elem.degree

        if cellname not in ("triangle", "quadrilateral"):
            raise NotImplementedError(
                f"Only 2D triangle or quadrilateral cells; got {cellname}"
            )
        if degree != 1:
            raise NotImplementedError(
                "This limiter is intended for degree=1 only."
            )

        self.mesh = V.mesh
        self.V_dg = V
        self.alpha = alpha

        # For DG(1) on triangles, each cell has 3 dofs; on quads, 4 dofs
        if cellname == "triangle":
            self.dofs_per_cell = 3
        elif cellname == 'quadrilateral':
            self.dofs_per_cell = 4
        else:
            raise NotImplementedError(
                f"Only 2D triangle or quadrilateral cells are supported; got {cellname}"
            )
        


            mesh_element_name = self.mesh.topology.cell_name()  
        elem_DG0 = basix.ufl.element(
            family="DG", cell=self.mesh.topology.cell_name(), degree=0
        )
        elem_CG1 = basix.ufl.element(
            family="CG", cell=self.mesh.topology.cell_name(), degree=1
        )
        # A DG(0) space for cellwise (centroid) averages
        self.V_p0 = fem.functionspace(mesh=self.mesh, element=elem_DG0)
        # A CG(1) space for per-vertex min/max
        self.V_cg = fem.functionspace(mesh=self.mesh, element=elem_CG1)

        # Store centroids, vertex-based bounds
        self.centroids = fem.Function(self.V_p0, name="DG0_centroids")
        self.min_field = fem.Function(self.V_cg, name="vertex_min")
        self.max_field = fem.Function(self.V_cg, name="vertex_max")

    def _update_centroids(self, field):
        
        solution = project(field, self.V_p0)
        
        self.centroids.x.array[:] = solution.x.array

    def compute_bounds(self, field):
        """
        For each vertex, find the min and max cell-average among adjacent cells.
        Store in self.min_field, self.max_field.
        """
        # 1) Compute cellwise centroids in self.centroids
        self._update_centroids(field)

        # 2) Initialize vertex-based min/max
        self.min_field.x.array[:] = 1.0e20
        self.max_field.x.array[:] = -1.0e20

        # 3) Build connectivity from vertices -> cells
        tdim = self.mesh.topology.dim
        self.mesh.topology.create_connectivity(0, tdim)
        v2c = self.mesh.topology.connectivity(0, tdim)
        if v2c is None:
            return

        offsets = v2c.offsets
        indices = v2c.array

        centroids_arr = self.centroids.x.array
        min_arr = self.min_field.x.array
        max_arr = self.max_field.x.array

        num_vertices_local = len(offsets) - 1
        for v_local in range(num_vertices_local):
            cell_start = offsets[v_local]
            cell_end   = offsets[v_local+1]
            local_min = 1.0e20
            local_max = -1.0e20
            for c_local in indices[cell_start:cell_end]:
                if c_local < len(centroids_arr):
                    val = centroids_arr[c_local]
                    local_min = min(local_min, val)
                    local_max = max(local_max, val)
            min_arr[v_local] = local_min
            max_arr[v_local] = local_max

        # Push updated min/max to ghost vertices in parallel
        self.min_field.x.scatter_forward()
        self.max_field.x.scatter_forward()

    def apply_limiter(self, field):
        """
        Perform slope limiting on the DG(1) field, ensuring dofs remain
        within [qmin, qmax] from the vertex-based patch.
        """
        dofmap = self.V_dg.dofmap
        field_arr = field.x.array
        centroids_arr = self.centroids.x.array

        min_arr = self.min_field.x.array
        max_arr = self.max_field.x.array

        num_cells_local = self.mesh.topology.index_map(
            self.mesh.topology.dim).size_local

        for c_local in range(num_cells_local):
            q_avg = centroids_arr[c_local]  # DG(0) average
            cdofs = dofmap.cell_dofs(c_local)
            qvals = field_arr[cdofs]

            # Collect the min/max from each corner dof's associated vertex
            qmin = 1.0e20
            qmax = -1.0e20
            for d in cdofs:
                if 0 <= d < len(min_arr):
                    qmin = min(qmin, min_arr[d])
                    qmax = max(qmax, max_arr[d])

            alpha = self.alpha
            for val in qvals:
                dq = val - q_avg
                if abs(dq) < 1e-14:
                    continue
                if dq > 0:
                    alpha = min(alpha, (qmax - q_avg)/dq)
                else:
                    alpha = min(alpha, (qmin - q_avg)/dq)

            # Scale dofs
            for i, d in enumerate(cdofs):
                dq = qvals[i] - q_avg
                field_arr[d] = q_avg + alpha*dq

        # Forward scatter in parallel
        field.x.scatter_forward()

    def apply(self, field):
        """
        Full procedure: compute vertex-based bounds, then limit.
        """
        assert field.function_space is self.V_dg, \
            "Field must be in the same DG(1)/DQ(1) space as this limiter."
        self.compute_bounds(field)
        self.apply_limiter(field)



if __name__ == "__main__":
    comm = MPI.COMM_WORLD

    # Choose a 2D mesh with triangles or quadrilaterals:
    # cell_type = dolfinx.cpp.mesh.CellType.triangle
    cell_type = dolfinx.cpp.mesh.CellType.quadrilateral

    domain = dmesh.create_unit_square(comm, 4, 4, cell_type=cell_type)

    # For triangles, use ("DG", 1); for quads, use ("DQ", 1):
    mesh_element_name = domain.topology.cell_name()  
    basis_functions_degree = 1
    finite_element = basix.ufl.element(
        family="DG", cell=domain.topology.cell_name(), degree=basis_functions_degree
    )
    V_dg = fem.functionspace(mesh=domain, element=finite_element)

    # Create a test field
    field = fem.Function(V_dg, name="field")

    # Example: add a Gaussian bump
    def bump(x):
        return np.exp(-50*((x[0] - 0.8)**2 + (x[1] - 0.8)**2))
    field.interpolate(bump)

    # Create and apply the limiter
    limiter = VertexBasedLimiter(V_dg)
    limiter.apply(field)

    if comm.rank == 0:
        print("Field has been limited on rank 0.")