        
import pyvista
import numpy as np
import dolfinx
from dolfinx.mesh import CellType
import os
import base64
from IPython.display import HTML

def mesh(functionspace, title, figurepath, figurename, show_edges=False):

    topology, cell_types, x = dolfinx.plot.vtk_mesh(functionspace)
    grid = pyvista.UnstructuredGrid(topology, cell_types, x)
    # Plotting
    plotter = pyvista.Plotter(off_screen=True)
    plotter.add_mesh(grid, show_edges=show_edges, cmap="coolwarm")
    plotter.add_title(title)
    plotter.camera_position = "xy"
    
    plotter.screenshot(f"{os.path.join(figurepath, figurename)}.png")
    print(f'image saved in {os.path.join(figurepath, figurename)}.png')

def scalar_field(function, title, figurepath, figurename, show_edges=False):

    functionspace = function.function_space
    mesh = functionspace.mesh
    element = functionspace.ufl_element()
    degree = element.degree


    if degree == 0:
        # dom = q_n.function_space.mesh
        cells, cell_types, x = dolfinx.plot.vtk_mesh(mesh)  # not functionspace
        grid = pyvista.UnstructuredGrid(cells, cell_types, x)
        grid.cell_data["value"] = function.x.array


    else:
        # Use standard plotting for non-DG0 spaces
        cells, cell_types, x = dolfinx.plot.vtk_mesh(functionspace)
        grid = pyvista.UnstructuredGrid(cells, cell_types, x)
        grid.point_data["value"] = function.x.array

    scalars = "value"

    # Plotting
    plotter = pyvista.Plotter(off_screen=True)
    plotter.add_mesh(grid, scalars=scalars, show_edges=show_edges, cmap="coolwarm")
    plotter.add_title(title)
    plotter.camera_position = "xy"
    
    plotter.screenshot(f"{os.path.join(figurepath, figurename)}.png")
    print(f'image saved in {os.path.join(figurepath, figurename)}.png')

            
def scalar_fields(functions, titles, figurepath, figurename, show_edges=False):

    # assume the inputs is a list of functoins from the same space
    function = functions[0]
    functionspace = function.function_space
    mesh = functionspace.mesh
    element = functionspace.ufl_element()
    degree = element.degree

    # Plotting
    plotter = pyvista.Plotter(shape=(1, len(functions)), off_screen=True)
    for i, (function, title) in enumerate(zip(functions, titles)):
        plotter.subplot(0, i)

        if degree == 0:
            # dom = q_n.function_space.mesh
            cells, cell_types, x = dolfinx.plot.vtk_mesh(mesh)  # not functionspace
            grid = pyvista.UnstructuredGrid(cells, cell_types, x)
            grid.cell_data["value"] = function.x.array


        else:
            # Use standard plotting for non-DG0 spaces
            cells, cell_types, x = dolfinx.plot.vtk_mesh(functionspace)
            grid = pyvista.UnstructuredGrid(cells, cell_types, x)
            grid.point_data["value"] = function.x.array

        scalars = "value"


        plotter.add_mesh(grid, scalars=scalars, show_edges=show_edges, cmap="coolwarm")
        plotter.add_title(title)
        plotter.camera_position = "xy"
        
    plotter.screenshot(f"{os.path.join(figurepath, figurename)}.png")
    print(f'image saved in {os.path.join(figurepath, figurename)}.png')

        
    
def DG_illustration():
    import matplotlib.pyplot as plt
    import numpy as np

    # Define cell edges
    cell_edges = np.array([0.0, 1.0, 2.0, 3.0])

    # DG-0: constant per cell
    dg0_values = [1.0, 2.0, 0.5]

    # DG-1: linear per cell (discontinuous at interfaces)
    dg1_left = [1.0, 2.0, 0.5]
    dg1_right = [1.5, 1.0, 1.0]

    # CG-1: shared nodes at cell interfaces
    cg_nodes = np.array([0.0, 1.0, 2.0, 3.0])
    cg_values = np.array([1.0, 1.5, 1.0, 1.2])  # arbitrary smooth-ish function

    # Create subplots
    fig, (ax0, ax1, ax2) = plt.subplots(1, 3, figsize=(15, 4), sharey=True)

    # Plot DG-0
    for i in range(3):
        x = [cell_edges[i], cell_edges[i+1]]
        y = [dg0_values[i], dg0_values[i]]
        ax0.plot(x, y, 'b', lw=2)


    for edge in cell_edges[1:-1]:
        ax0.plot([edge, edge], [0, 2.5], 'k--', alpha=0.5)

    ax0.set_title("DG-0 (Piecewise Constant)")
    ax0.set_xlabel("x")
    ax0.set_ylabel("u(x)")
    ax0.set_xlim(0, 3)
    ax0.set_ylim(0, 2.5)

    # Plot DG-1
    for i in range(3):
        x = np.linspace(cell_edges[i], cell_edges[i+1], 10)
        y = np.linspace(dg1_left[i], dg1_right[i], 10)
        ax1.plot(x, y, 'r', lw=2)

    for edge in cell_edges[1:-1]:
        ax1.plot([edge, edge], [0, 2.5], 'k--', alpha=0.5)

    ax1.set_title("DG-1 (Piecewise Linear)")
    ax1.set_xlabel("x")
    ax1.set_xlim(0, 3)

    # Plot CG-1
    ax2.plot(cg_nodes, cg_values, 'g', lw=2)
    ax2.set_title("CG-1 (Continuous Linear)")
    ax2.set_xlabel("x")
    ax2.set_xlim(0, 3)
    
    for edge in cell_edges[1:-1]:
        ax2.plot([edge, edge], [0, 2.5], 'k--', alpha=0.5)

    plt.tight_layout()
    return fig

import matplotlib.pyplot as plt
import numpy as np

def DG_facet_integral_illustration():
    # Define two cells
    cell_edges = np.array([0.0, 1.0, 2.0])

    # Define linear values in each cell: left value and right value
    q_left_cell = (0.0, 1.0)   # from x=0 to x=1
    q_right_cell = (2.0, 1.0)  # from x=1 to x=2

    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 3), sharey=True)

    for ax, normal_side in zip((ax1, ax2), ('+', '-')):

        # Plot left cell
        x_left = np.linspace(cell_edges[0], cell_edges[1], 10)
        y_left = np.linspace(q_left_cell[0], q_left_cell[1], 10)
        ax.plot(x_left, y_left, 'r', lw=2)

        # Plot right cell
        x_right = np.linspace(cell_edges[1], cell_edges[2], 10)
        y_right = np.linspace(q_right_cell[0], q_right_cell[1], 10)
        ax.plot(x_right, y_right, 'r', lw=2)

        # Vertical interface line
        ax.axvline(x=1.0, color='blue', linestyle='--', lw=2)

        # Add q^+ and q^- labels
        ax.text(1.0 - 0.05, q_left_cell[1] + 0.1, r'$q^+$', ha='right', va='bottom', fontsize=13)
        ax.text(1.0 + 0.05, q_right_cell[0] + 0.1, r'$q^-$', ha='left', va='bottom', fontsize=13)

        # Add normal arrow
        if normal_side == '+':
            ax.annotate(r'$\mathbf{n}^+$',
                        xy=(1.0, 1.5), xytext=(1.2, 1.5),
                        arrowprops=dict(arrowstyle="<-", color='black', lw=1.5),
                        fontsize=13, ha='left', va='center')
            ax.set_title("Outward normal $\\mathbf{n}^+$ (left cell)", fontsize=12)
        else:
            ax.annotate(r'$\mathbf{n}^-$',
                        xy=(1.0, 1.5), xytext=(0.8, 1.5),
                        arrowprops=dict(arrowstyle="<-", color='black', lw=1.5),
                        fontsize=13, ha='right', va='center')
            ax.set_title("Outward normal $\\mathbf{n}^-$ (right cell)", fontsize=12)

        # Common formatting
        ax.set_xlabel("x")
        ax.set_xlim(0, 2)
        ax.set_ylim(0, 2.5)

    ax1.set_ylabel("q(x)")
    plt.tight_layout()
    return fig


def display_image(image_path=None):
    if image_path is None:
        print("No image path provided.")
    else:
        with open(f"{image_path}", "rb") as image_file:
            encoded_string = base64.b64encode(image_file.read()).decode("utf-8")
            return HTML(f'<img src="data:image/png;base64,{encoded_string}" width="600" />')
